from collections import OrderedDict
import numpy as np

def read_fasta_to_dic(filename):
    """
    function used to parser small fasta
    still effective for genome level file
    """
    fa_dic = OrderedDict()

    with open(filename, "r") as f:
        for n, line in enumerate(f.readlines()):
            if line.startswith(">"):
                if n > 0:
                    fa_dic[short_name] = "".join(seq_l)  # store previous one

                full_name = line.strip().replace(">", "")
                short_name = full_name.split(" ")[0]
                seq_l = []
            else:  # collect the seq lines
                if len(line) > 8:  # min for fasta file is usually larger than 8
                    seq_line1 = line.strip()
                    seq_l.append(seq_line1)

        fa_dic[short_name] = "".join(seq_l)  # store the last one
    return fa_dic

def reverse_fasta(ref):
    base_dict={'A':'T',"T":"A","C":"G","G":"C"}
    ref=np.array(list(ref.upper()))
    for index,element in enumerate(ref):
        ref[index] = base_dict[element]
    reverse_fasta = list(reversed(''.join(ref)))
    return ''.join(reverse_fasta)

def save_fasta_dict(fasta_dict,path):
    f=open(path,'w+')
    for key,value in fasta_dict.items():
        f.write('>'+key+'\n')
        line = len(value) //70 + 1
        for i in range(0, line):
            f.write(value[i*70:(i+1)*70]+'\n')
        f.close()
# fasta=read_fasta_to_dic("/data/Ecoli_23s/23S_rRNA.fasta")
# for key,value in fasta.items():
#     fasta[key]=reverse_fasta(value)
# save_fasta_dict(fasta,'test.fasta')


#!/usr/bin/env python
import pandas as pd
import os
import numpy as np
from tqdm import tqdm
import pysam
import argparse
import shutil
import time

def create_result_folder(output_path):
    print("Try to build generate the result path: "+output_path)
    if os.path.exists(output_path):
        print("Folder exists. It will be overwrite after 5 secs ...")
        time.sleep(5)
        print("Deleting folder...")
        shutil.rmtree(output_path)
        print("Folder deleted. Creating folder...")
        os.mkdir(output_path)
        print("Folder created.")
    else:
        print("Creating output folder...")
        os.mkdir(output_path)
        print("Folder created.")

def select_sub_reads(line):
    result = line[0].split("/")[-1].split(".")[0]
    return result

def copy2new_path(line,results_path):
    global pbar
    fold_name = line[0].split('/')[-2]
    if not os.path.exists(results_path+'/single/'+fold_name):
        os.mkdir(results_path+'/single/'+fold_name)
    os.system('cp ' + line[0] + ' ' + results_path+'/single/'+fold_name)
    pbar.update(1)

def transfer_fast5(total_fl, df,results_path):
    global pbar
    total_fl.columns=[0,'readname']
    df = df[['readname']]
    df[2] = '.'

    temp = pd.merge(total_fl, df, on='readname', how='inner')

    temp=temp.drop_duplicates(subset='readname')
    print("Start to extract "+str(temp.shape[0])+' reads ...')
    # temp=temp.sample(n=1544)
    pbar = tqdm(total=temp.shape[0], position=0, leave=True)


    os.mkdir(results_path+'/single/')
    temp.apply(copy2new_path,results_path=results_path, axis=1)
    pbar.close()

def main(args):
    fast5_path = args.fast5
    results_path = args.output
    sam_file=args.bam
    create_result_folder(results_path)
    print("Loading bam files ... ")
    if sam_file.split(".")[-1] == 'bam':
        sam_file=pysam.AlignmentFile(sam_file,'rb')
    else:
        sam_file=pysam.AlignmentFile(sam_file,'r')

    read_list=[]
    if args.chrom is None or args.pos is None:
        for read in sam_file.fetch():
            read_list.append(read.qname)
    else:
        for read in sam_file.fetch(args.chrom, args.pos - 10, args.pos + 10):
            read_list.append(read.qname)
    df = pd.DataFrame(read_list)
    df.columns=['readname']

    #READ FAST5 FILE LIST
    print("Collecting fast5 files ... ")
    os.system("find " + fast5_path + " -name \"*.fast5\" >" + results_path+"/files.txt")
    fast5_file =  results_path+"/files.txt"
    print("Generated fast5 list file: ", fast5_file)

    print("Subsampling fast5 files ... ")
    total_fl = []
    for i in open(fast5_file, "r"):
        total_fl.append(i.rstrip())
    total_fl = pd.DataFrame(np.array(total_fl))
    total_fl[1] = total_fl.apply(select_sub_reads, axis=1)
    transfer_fast5(total_fl, df, results_path)

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("-f","--fast5", required=True,
                        help="fast5_path")
    parser.add_argument('-b',"--bam", required=True,
                        help="bam_path")
    parser.add_argument('-o',"--output", default='subsample_single', help="output_file")
    parser.add_argument("--chrom",help="Gene or chromosome name(head of your fasta file)")
    parser.add_argument("--pos", type=int, help="site of your interest")
    # parser.add_argument("-f","--fast5", default='/data/Ecoli_23s/data/L_rep2/single/',
    #                     help="fast5_path")
    # parser.add_argument('-b',"--bam", default="/data/Ecoli_23s/data/L_rep2/file.bam",
    #                     help="bam_path")
    # parser.add_argument('-o',"--output", default='/data/Ecoli_23s/data/L_rep2/subsample2', help="output_file")
    # parser.add_argument("--chrom", default='NR_103073.1',help="Gene or chromosome name(head of your fasta file)")
    # parser.add_argument("--pos", default=2029, type=int, help="site of your interest")
    args = parser.parse_args()
    main(args)



import pandas as pd
import plotnine as p9
import numpy as np
from matplotlib import pyplot as plt

plt.rcParams['pdf.fonttype'] = 42
# plt.rcParams['font.sans-serif'] = ['Arial']
SIG_PCTL_RANGE = (2.5, 97.5)


# def draw_boxplot(df,results_path,pos,base_list,title):
#     item_list = ['Mean', 'STD', 'Median', 'Dwell_time']
#     plot_list=[]
#     for item in item_list:
#
#         sig_min, sig_max = np.percentile(df[item], SIG_PCTL_RANGE)
#         sig_diff = sig_max - sig_min
#         ylim_tuple = (sig_min - sig_diff * 0.1, sig_max + sig_diff * 0.1)
#
#         plot = p9.ggplot(df, p9.aes(x='position', y=item, fill='type')) \
#                + p9.theme_bw() \
#                +p9.geom_boxplot( outlier_shape='',position=p9.position_dodge(0.9),size=0.25) \
#                + p9.scale_fill_manual(values={"Sample": "#ff6f91", "Control": "#7389af"}) \
#                + p9.scale_x_discrete(labels=list(base_list)) \
#                + p9.theme(
#                     figure_size=(6, 3),
#                     panel_grid_minor=p9.element_blank(),
#                     axis_text=p9.element_text(size=13),
#                     axis_title=p9.element_text(size=13),
#                     title=p9.element_text(size=13),
#                     legend_position='none')\
#                + p9.ylim(ylim_tuple)\
#                + p9.labs(title=title, x=str(pos + 1), y=item)
#
#         plot.save(filename=results_path + "/" + item + "_boxplot.pdf", dpi=300)
#
# def draw_violin(df,results_path,pos,base_list,title):
#
#     item_list = ['Mean', 'STD', 'Median', 'Dwell_time']
#     for item in item_list:
#
#         sig_min, sig_max = np.percentile(df[item], SIG_PCTL_RANGE)
#         sig_diff = sig_max - sig_min
#         ylim_tuple = (sig_min - sig_diff * 0.1, sig_max + sig_diff * 0.1)
#
#         plot = p9.ggplot(df, p9.aes(x='position', y=item, fill='type')) \
#                + p9.geom_violin(style='left-right',position=p9.position_dodge(0),color='none',width=1.5) \
#                + p9.theme_bw() \
#                + p9.scale_fill_manual(values={"Sample": "#ff6f91", "Control": "#7389af"}) \
#                + p9.scale_x_discrete(labels=list(base_list)) \
#                + p9.theme(
#             figure_size=(6, 3),
#             panel_grid_minor=p9.element_blank(),
#             axis_text=p9.element_text(size=13),
#             axis_title=p9.element_text(size=13),
#             title=p9.element_text(size=13),
#             legend_position='none'
#         )\
#         + p9.ylim(ylim_tuple)
#
#         plot = plot + p9.labs(title=title, x=str(pos + 1), y=item)
#         # plot.render_matplotlib()
#         plot.save(filename=results_path + "/" + item + "_violin.pdf", dpi=300)
# print(plot)

def signal_plot(df, results_path, pos, base_list, title, plot_type,filter=False):
    item_list = ['Mean', 'STD', 'Median', 'Dwell time']


    if plot_type != 'merged':

        for item in item_list:
            sig_min, sig_max = np.percentile(df[item], SIG_PCTL_RANGE)
            sig_diff = sig_max - sig_min
            ylim_tuple = (sig_min - sig_diff * 0.1, sig_max + sig_diff * 0.1)

            plot = p9.ggplot(df, p9.aes(x='position', y=item, fill='type')) \
                   + p9.theme_bw() \
                   + p9.scale_fill_manual(values={"Sample": "#ff6f91", "Control": "#7389af"}) \
                   + p9.scale_x_discrete(labels=list(base_list)) \
                   + p9.theme(
                figure_size=(6, 3),
                panel_grid_minor=p9.element_blank(),
                axis_text=p9.element_text(size=13),
                axis_title=p9.element_text(size=13),
                title=p9.element_text(size=13),
                legend_position='none'
            ) \
                   + p9.ylim(ylim_tuple)
            plot = plot + p9.labs(title=title, x=str(pos + 1), y=item)
            if plot_type == 'boxplot':
                plot = plot + p9.geom_boxplot(outlier_shape='', position=p9.position_dodge(0.9), size=0.25)
            elif plot_type == 'violin_plot':
                plot = plot + p9.geom_violin(style='left-right', position=p9.position_dodge(0), color='none', width=1.5)
            else:
                raise Exception("Unsupported figure type!")
            # plot.render_matplotlib()
            if item == 'Dwell time':
                item = 'Dwell_time'
            plot.save(filename=results_path + "/" + item + "_" + plot_type + ".pdf", dpi=300)
    else:
        new_df = None
        for item in item_list:
            # collect data
            temp = df[[item, 'position', 'type']].copy()
            temp.columns = ['value', 'position', 'Group']
            temp.loc[:, 'stats'] = item
            if filter:
                sig_min, sig_max = np.percentile(temp['value'], SIG_PCTL_RANGE)
                sig_diff = sig_max - sig_min
                ylim_tuple = [sig_min - sig_diff * 0.1, sig_max + sig_diff * 0.1]
                temp = temp[(temp['value'] >= ylim_tuple[0]) & (temp['value'] <= ylim_tuple[1])]
            if new_df is None:
                new_df = temp
            else:
                new_df = pd.concat([new_df, temp], axis=0)

        plot = p9.ggplot(new_df, p9.aes(x='position', y="value", fill='Group')) \
               + p9.theme_bw() \
               + p9.scale_fill_manual(values={"Sample": "#ff6f91", "Control": "#7389af", "Single": "#a3abbd"}) \
               + p9.scale_x_discrete(labels=list(base_list)) \
               + p9.theme(
            figure_size=(8, 8),
            panel_grid_minor=p9.element_blank(),
            axis_text=p9.element_text(size=13),
            axis_title=p9.element_text(size=13),
            title=p9.element_text(size=13),
            legend_position='bottom',
            legend_title=p9.element_blank(),
            strip_text=p9.element_text(size=13),
            strip_background=p9.element_rect(alpha=0),

        ) \
               + p9.facet_grid('stats ~', scales='free_y')
        plot = plot + p9.labs(title=title, x=str(pos + 1), y='')

        if new_df['Group'].drop_duplicates().shape[0] == 1:
            plot2 = plot + p9.geom_violin(color='none', position=p9.position_dodge(0.9), width=1)
            plot2 = plot2 + p9.geom_boxplot(outlier_shape='', position=p9.position_dodge(0.9), size=0.2, width=0.1)
            plot2.save(filename=results_path + "/Merged_single.pdf", dpi=300)
        else:
            plot1 = plot + p9.geom_boxplot(outlier_shape='', position=p9.position_dodge(0.9), size=0.2, width=0.75,alpha = 0.8)
            plot1.save(filename=results_path + "/Merged_boxplot.pdf", dpi=300)
            plot2 = plot + p9.geom_violin(style='left-right', position=p9.position_dodge(0), color='none', width=1.5,alpha = 0.8)
            plot2.save(filename=results_path + "/Merged_violin.pdf", dpi=300)


def draw_signal(df, start, base,start_index,end_index):
    df = pd.DataFrame(df)
    df.columns = ['raw']
    start=start[start_index:end_index]
    df=df[start[0]:start[-1]]
    df = df.reset_index()
    plot = p9.ggplot(df, p9.aes(x='index', y="raw")) \
           + p9.theme_bw() \
           + p9.geom_line() \
           + p9.theme(
        figure_size=(6, 3),
        panel_grid_minor=p9.element_blank(),
        axis_text=p9.element_text(size=13),
        axis_title=p9.element_text(size=13),
        title=p9.element_text(size=13),
        strip_text=p9.element_text(size=13),
        legend_position='none',
        strip_background=p9.element_rect(alpha=0)
    )
    # plot.save(filename="/home/zhguo/Dropbox/@labfiles/projects/GUO/ONT_showcase_tool/plot/signal.pdf", dpi=300)
    for item in start:
        plot = plot + p9.geom_vline(xintercept=item, linetype='dashed', color='red')
    print(plot)
    # plot.save(filename="/home/zhguo/Dropbox/@labfiles/projects/GUO/ONT_showcase_tool/plot/norm_signal_aligned.pdf", dpi=300)

    # print(1)



#!/usr/bin/env python
import argparse
import os
from nanoCEM.cem_utils import read_fasta_to_dic,reverse_fasta
import pandas as pd
from nanoCEM.plot import signal_plot
from plotnine.exceptions import PlotnineWarning
import warnings
import time
warnings.filterwarnings("ignore", category=PlotnineWarning)
import numpy as np
def init_parser():
    def add_public_argument(parser_input):
        parser_input.add_argument("--chrom", required=True, help="Gene or chromosome name(head of your fasta file)")
        parser_input.add_argument("--pos", required=True, type=int, help="site of your interest")
        parser_input.add_argument("--len", default=10, type=int, help="region around the position")
        parser_input.add_argument("--strand", default="+", help="Strand of your interest")
        parser_input.add_argument("--ref", required=True, help="fasta file")
        parser_input.add_argument('--norm', action='store_true', help='Turn on the normalization mode')
        parser_input.add_argument("--overplot-number", default=500, type=int,
                                  help="Number of read will be used to plot")
        parser_input.add_argument('--rna', action='store_true', help='Turn on the RNA mode')
    # Define the argument parser
    parser = argparse.ArgumentParser(description='A sample tool designed to visualize the features that distinguish between two groups of ONT data at the site level. It supports two re-squiggle pipeline(Tombo and f5c).')
    subparsers = parser.add_subparsers(dest='function')

    # tombo subparser
    parser_tombo = subparsers.add_parser('tombo', help='tackle tombo re-squiggle')
    parser_tombo.add_argument('--basecall_group', default="RawGenomeCorrected_000",
                        help='The attribute group to extract the training data from. e.g. RawGenomeCorrected_000')
    parser_tombo.add_argument('--basecall_subgroup', default='BaseCalled_template',
                        help='Basecall subgroup Nanoraw resquiggle into. Default is BaseCalled_template')
    parser_tombo.add_argument('-i', "--input_fast5", required=True,
                        help="input_fast5_file")
    parser_tombo.add_argument('-c', "--control_fast5",
                        help="control_fast5_file")
    parser_tombo.add_argument('-o', "--output", default="tombo_result", help="output_file")
    parser_tombo.add_argument('-t', "--cpu", default=4, type=int, help="num of process")
    add_public_argument(parser_tombo)

    # f5c subparser
    parser_f5c = subparsers.add_parser('f5c', help='tackle f5c re-squiggle')
    parser_f5c.add_argument("-i", "--input", required=True,
                        help="blow5_path")
    parser_f5c.add_argument('-c', "--control",
                        help="control_blow5_path")
    parser_f5c.add_argument('-o', "--output", default="f5c_result", help="output_file")

    parser_f5c.add_argument('--base_shift',type=int, default=0, help="output_file")
    add_public_argument(parser_f5c)
    return parser

if __name__ == '__main__':
    # Parse the arguments
    parser = init_parser()
    args = parser.parse_args()
    aligned_num_wt = 0
    aligned_num_ivt = 0
    # read reference
    # print(args)
    if args.function is None:
        raise RuntimeError("Please choose function from f5c or tombo, or use -h to view the help document")
    subsample_num = args.overplot_number
    fasta = read_fasta_to_dic(args.ref)
    # length filter
    args.pos = args.pos - 1
    length_gene = len(fasta[args.chrom])
    if args.pos + args.len + 1 >= length_gene or args.pos - args.len <= 0:
        raise RuntimeError("The position requested is too close to the border (pos-len>0 and pos+len<length of fasta)")
    base_list = fasta[args.chrom][args.pos - args.len:args.pos + args.len + 1].upper()

    if args.strand == '-':
        base_list = "".join(list(reversed(base_list)))
        base_list = reverse_fasta(base_list)

    title = args.chrom + ':' + str(args.pos - args.len + 1) + '-' + str(args.pos + args.len + 2) + ':' + args.strand
    if args.rna:
        base_list = base_list.replace("T", "U")
        print("Running RNA mode ...")
        title ='RNA  ' + title
    else:
        print("Running DNA mode ...")
        title = 'DNA  ' + title

    results_path = args.output
    if not os.path.exists(results_path):
        os.mkdir(results_path)
    else:
        print("Output file existed! It will be overwrite after 5 secs ...")
        time.sleep(5)
        print("Continue ...")


    if args.function == 'tombo' :

        from  nanoCEM.read_tombo_resquiggle import create_read_list_file,extract_group
        wt_file = create_read_list_file(args.input_fast5, results_path)
        df_wt, aligned_num_wt = extract_group(args, wt_file, subsample_num)
        df_wt['type'] = 'Sample'
        try:
            ivt_file = create_read_list_file(args.control_fast5, results_path)
            df_ivt, aligned_num_ivt = extract_group(args, ivt_file, subsample_num)
            df_ivt['type'] = 'Control'
            df = pd.concat([df_wt, df_ivt])
            title = title + '   Sample:' + str(aligned_num_wt) + '  Control:' + str(aligned_num_ivt)
            category = pd.api.types.CategoricalDtype(categories=['Sample', "Control"], ordered=True)
            df['type'] = df['type'].astype(category)
        except:
            args.control_fast5 = None
        if args.control_fast5 is None:
            df = df_wt
            df_wt['type'] = 'Single'
            title = title + '   Sample:' + str(aligned_num_wt)
    elif args.function == 'f5c':
        from nanoCEM.read_f5c_resquiggle import read_blow5
        if args.base_shift < 0:
            raise RuntimeError("base_shift should not less than 0")
        # if (args.rna and args.strand=='+') or (not args.rna and args.strand=='-'):
        #     args.pos = args.pos + args.base_shift
        # else:
        #     args.pos = args.pos - args.base_shift
        df_wt, aligned_num_wt,nucleotide_type = read_blow5(args.input, args.pos, args.len, args.chrom, args.strand,subsample_num,args.base_shift,args.norm)
        df_wt['type'] = 'Sample'
        if nucleotide_type=='RNA' and not args.rna:
            raise RuntimeError("You need to add --rna to turn on the rna mode")
        try:
            df_ivt, aligned_num_ivt,_ = read_blow5(args.control, args.pos, args.len, args.chrom, args.strand,
                                                 subsample_num,args.base_shift,args.norm)
            df_ivt['type'] = 'Control'

            df = pd.concat([df_wt, df_ivt])
            category = pd.api.types.CategoricalDtype(categories=['Sample', "Control"], ordered=True)
            df['type'] = df['type'].astype(category)

            title = title + '   Sample:' + str(aligned_num_wt) + '  Control:' + str(aligned_num_ivt)
        except:
            args.control = None
        if args.control is None:
            df = df_wt
            df_wt['type'] = 'Single'
            title = title + '   Sample:' + str(aligned_num_wt)


    category_data = [str(args.pos + x) for x in range(-args.len, args.len + 1)]
    category = pd.api.types.CategoricalDtype(categories=category_data, ordered=True)
    df['position'] = df['position'].astype(category)


    # draw_volin(df,results_path,args.pos,base_list,title)
    # draw_boxplot(df,results_path,args.pos,base_list,title)
    print("Start to generate plots and save  in "+ results_path)
    # if args.function == 'f5c':
    #     if (args.rna and args.strand == '+') or (not args.rna and args.strand == '-'):
    #         args.pos = args.pos - args.base_shift
    #     else:
    #         args.pos = args.pos + args.base_shift
    percentile_filter=False
    if aligned_num_wt > 50 and aligned_num_ivt > 50:
        percentile_filter=True

    signal_plot(df, results_path, args.pos, base_list, title, 'merged',percentile_filter)
    # signal_plot(df, results_path, args.pos, base_list, title, 'boxplot')
    # signal_plot(df, results_path, args.pos, base_list, title, 'violin_plot')
    print('finished')



import re
import numpy as np
import pandas as pd
import pyslow5
import pysam
from tqdm import tqdm
from nanoCEM.normalization import normalize_signal,normalize_signal_with_lim
# from nanoCEM.plot import draw_signal
import os
import argparse
score_dict={}
nucleotide_type=None
def extract_feature(line,strand,base_shift=2,norm=True):
    global nucleotide_type
    pbar.update(1)
    read_id = line[0]
    # if read_id !='562eeb47-2b86-4fc7-abfc-5dce62f511ed':
    #     return None
    if read_id not in info_dict:
        return None
    # tackle moves tag
    moves_string = line[14]
    moves_string = re.sub('ss:Z:', '', moves_string)
    moves_string = re.sub('D', 'D,', moves_string)
    moves_string = re.sub('I', 'I,', moves_string)
    # print(moves_string)
    moves = re.split(r',+', moves_string)
    moves = moves[:-1]
    # extract index and generate event_length and event_start
    insertion = 0
    event_length = []
    for i,item in enumerate(moves):
        if 'D' in item:
            deletion = int(item[:-1])
            for i in range(deletion):
                event_length.append(0)
        elif 'I' in item:
            if i == 0 :
                continue
            else:
                return None
        elif '=' in item:
            return None
        else:
            event_length.append(int(item))
    # build event_length from move table
    read = s5.get_read(read_id, aux=["read_number", "start_mux"],pA=True)
    start_index = line[2]
    end_index = line[3]
    event_length = np.array(event_length)

    # identify RNA or DNA
    if nucleotide_type is None:
        if line[7]>line[8]:
            nucleotide_type='RNA'
        else:
            nucleotide_type='DNA'
    #  assert len_raw_signal in paf and blow5
    try:
        assert end_index-start_index == np.sum(event_length)
        assert read['len_raw_signal'] == line[1]
    except Exception:
        print("Warning: 1 read's length of signal is not equal between blow5 and paf")
        return None

    signal = read['signal']

    # create event start and flip all table
    signal = signal[start_index:end_index]
    if nucleotide_type=='RNA':
        signal = np.flip(signal)
        event_length = np.flip(event_length)

    event_starts = event_length.cumsum()
    event_starts = np.insert(event_starts, 0, 0)[:-1]
    if norm:
        signal = normalize_signal_with_lim(signal)

    # filter too short or long dwell time
    # dwell_filter_pctls = (0.5, 99.5)
    # dwell_min, dwell_max = np.percentile(event_length, dwell_filter_pctls)
    # valid_bases = np.logical_and.reduce(
    #     (
    #         event_length > dwell_min,
    #         event_length < dwell_max,
    #         np.logical_not(np.isnan(event_length)),
    #     )
    # )
    # event_length[~valid_bases] = 0

    # normalized signal
    # draw_signal(signal, event_starts,None, 8850,8855)


    # index query and reference
    aligned_pair=info_dict[read_id]['pairs']
    qlen = info_dict[read_id]['query_length']
    # correct index about DNA and RNA
    if nucleotide_type == 'RNA':
        if strand == '+':
            gap = qlen - event_length.shape[0]
            aligned_pair[0] = aligned_pair[0] - gap
            aligned_pair = aligned_pair[aligned_pair[0] >= 0]
        elif strand == '-':
            aligned_pair[0] = event_length.shape[0]-aligned_pair[0]-1
    else:
        if strand == '-':
            aligned_pair[0] = qlen - aligned_pair[0] - 1
        aligned_pair=aligned_pair[aligned_pair[0] < event_length.shape[0]]
    # base shift
    if nucleotide_type == 'RNA':
        aligned_pair[0]=aligned_pair[0].values + base_shift
        aligned_pair = aligned_pair[aligned_pair[0] < event_length.shape[0]]
    else:
        aligned_pair[0] = aligned_pair[0].values - base_shift
        aligned_pair = aligned_pair[aligned_pair[0] >= 0]

    if aligned_pair.shape[0]==0:
        return None
    read_pos = aligned_pair[0].values
    ref_pos = aligned_pair[1].values

    # extract raw signal by event length and event start
    total_feature_per_reads = []
    try:
        raw_signal_every = [signal[event_starts[x]:event_starts[x] + event_length[x]] for x in
                            read_pos]
        # if aligned_pair.shape[0] < 11:
        #     draw_signal(signal[event_starts[read_pos[0]]:event_starts[read_pos[-1]+1]], event_starts[read_pos[0]:read_pos[-1]+1]-event_starts[read_pos[0]], base_list)
    except Exception:
        print(1)
    # calculate mean median and dwell time
    for i, element in enumerate(raw_signal_every):
        if event_length[read_pos[i]] == 0:
            continue
        temp = [np.mean(element), np.std(element), np.median(element), event_length[read_pos[i]],ref_pos[i]]
        total_feature_per_reads.append(temp)
    return total_feature_per_reads

def extract_pairs_pos(bam_file,position,length,chromosome,strand):

    result_dict={}
    for read in bam_file.fetch(chromosome,position-length, position+length+1):
        if strand == '+' and read.is_reverse:
            continue
        if strand == '-' and not read.is_reverse:
            continue
        # if read.qname == '7dcec2bc-65f3-41fe-8981-dd3af9fa6e67':
        #     print(1)
        start_position=read.reference_start
        end_position=read.reference_end
        if position < start_position or position > end_position:
            continue
        # unit
        aligned_pair = np.array(read.aligned_pairs)
        aligned_pair = pd.DataFrame(aligned_pair)
        if strand == '-':
            aligned_pair[0] = aligned_pair[0].iloc[::-1]
        aligned_pair.dropna(inplace=True,ignore_index=True)
        aligned_pair = aligned_pair[(aligned_pair[1] >= position - length) & (aligned_pair[1] <= position + length)]

        temp={}
        temp['pairs']=aligned_pair
        temp['query_length'] = read.query_length
        result_dict[read.qname] = temp
    return result_dict



def read_blow5(path,position,length,chromo,strand,subsapmle_num=500,base_shift=2,norm=True):
    global info_dict,s5,pbar
    bam_file = path+".bam"
    if not os.path.exists(path+'.bam'):
        raise RuntimeError(path+'.bam' +"is not exist in your path!")
    if not os.path.exists(path+'.paf'):
        raise RuntimeError(path+'.paf' +"is not exist in your path!")
    if not os.path.exists(path+'.blow5'):
        raise RuntimeError(path+'.blow5' +"is not exist in your path!")
    bam_file=pysam.AlignmentFile(bam_file,'rb')
    # if rna_mode:
    #     if strand =='+':
    #         position=position+ (kmer_model-1)
    #     else:
    #         position=position- (kmer_model-1)
    info_dict=extract_pairs_pos(bam_file,position,length,chromo,strand)
    if info_dict == {}:
        raise RuntimeError("There is no read aligned on this position")
    info_df = pd.DataFrame(list(info_dict.keys()))
    slow5 = path+".blow5"
    s5 = pyslow5.Open(slow5, 'r')

    df=pd.read_csv(path+".paf",sep='\t',header=None)
    df=pd.merge(df,info_df,how='inner',on=0)
    if df.shape[0] == 0:
        raise RuntimeError("cannot found the record from bam in your paf file. Please check your f5c command ... ")
    if df.shape[0] / info_df.shape[0] < 0.8:
        print('There are '+str(info_df.shape[0]-df.shape[0])+" reads not found in your paf file ...")
    pbar = tqdm(total=df.shape[0], position=0, leave=True)
    df["feature"] = df.apply(extract_feature,strand=strand,base_shift=base_shift,norm=norm,axis=1)
    pbar.close()
    df.dropna(inplace=True)
    num_aligned = df.shape[0]
    if subsapmle_num < df.shape[0]:
        df=df.sample(n=subsapmle_num)
    final_feature=[]
    for item in df["feature"]:
        final_feature.extend(item)
    final_feature=pd.DataFrame(final_feature)
    final_feature.columns=['Mean','STD','Median','Dwell time','position']
    # if rna_mode:
    #     if strand == '+':
    #         final_feature['position']=final_feature['position'] - (kmer_model-1)
    #     else:
    #         final_feature['position']=final_feature['position'] + (kmer_model-1)

    final_feature['position'] = final_feature['position'].astype(int).astype(str)
    print('Extracted ', num_aligned, ' aligned reads from blow5 files')

    if num_aligned>50:
        dwell_filter_pctls = (0.5, 99.5)
        dwell_min, dwell_max = np.percentile(final_feature['Dwell time'].values, dwell_filter_pctls)
        final_feature = final_feature[(final_feature['Dwell time'] > dwell_min) & (final_feature['Dwell time'] < dwell_max)]

    return final_feature,num_aligned,nucleotide_type

# if __name__ == '__main__':
#     parser = argparse.ArgumentParser()
#     parser.add_argument("-i","--input", default='/data/Ecoli_23s/data/L_rep2/file',
#                         help="blow5_path")
#     parser.add_argument('-c',"--control", default='/data/Ecoli_23s/data/IVT_negative/file',
#                         help="control_blow5_path")
#     parser.add_argument('-o',"--output", default="/data/Ecoli_23s/f5c_results_2030_plus", help="output_file")
#     parser.add_argument("--chrom", default='NR_103073.1',help="Gene or chromosome name(head of your fasta file)")
#     parser.add_argument("--pos", default=2030, type=int, help="site of your interest")
#     parser.add_argument("--len", default=5, type=int, help="region around the position")
#     parser.add_argument("--strand", default="+", help="Strand of your interest")
#     parser.add_argument("--ref", default="/data/Ecoli_23s/23S_rRNA.fasta", help="fasta file")
#     parser.add_argument("--overplot-number", default=500, type=int, help="Number of read will be used to plot")
#     args = parser.parse_args()
#     args.pos = args.pos - 1
#     fasta=read_fasta_to_dic(args.ref)
#     base_list = fasta[args.chrom][args.pos-args.len:args.pos+args.len+1]
#     if args.strand == '-':
#         base_list="".join(list(reversed(base_list)))
#         base_list=reverse_fasta(base_list)
#     results_path = args.output
#     if not os.path.exists(results_path):
#         os.mkdir(results_path)
#     subsample_num = args.overplot_number
#
#     title = args.chrom + ':' + str(args.pos - args.len + 1) + '-' + str(args.pos + args.len + 2) + ':' + args.strand
#     df_wt,aligned_num_wt=read_blow5(args.input,args.pos,args.len,args.chrom,args.strand,subsample_num)
#
#     df_wt['type']='Sample'
#     try:
#         df_ivt,aligned_num_ivt=read_blow5(args.control,args.pos ,args.len,args.chrom,args.strand,subsample_num)
#         df_ivt['type'] = 'Control'
#
#         df=pd.concat([df_wt,df_ivt])
#         category = pd.api.types.CategoricalDtype(categories=['Sample',"Control"], ordered=True)
#         df['type'] = df['type'].astype(category)
#
#         title = title + '   Sample:' + str(aligned_num_wt) + '  Control:' + str(aligned_num_ivt)
#     except:
#         args.control=None
#     if args.control is None:
#         df = df_wt
#         df_wt['type'] = 'Single'
#         title = title + '   Sample:' + str(aligned_num_wt)
#
#     category_data = [str(args.pos + x) for x in range(-args.len, args.len + 1)]
#     category = pd.api.types.CategoricalDtype(categories=category_data, ordered=True)
#     df['position'] = df['position'].astype(category)
#
#
#     # df['Dwell_time'] = np.log10(df['Dwell_time'].values)
#
#     signal_plot(df, results_path, args.pos, base_list, title, 'merged')
#     signal_plot(df, results_path, args.pos, base_list, title,'boxplot')
#     signal_plot(df, results_path, args.pos, base_list, title, 'violin_plot')
#     print('\nsaved as ', args.output)



import numpy as np
import h5py
import pandas as pd
import plotnine as p9
import argparse
import os
import multiprocessing
from tqdm import tqdm
import random
# from ._stats import c_new_mean_stds
from nanoCEM.normalization import normalize_signal,normalize_signal_with_lim



def matrix_append(total_matrix,new_matrix,index):
    if new_matrix.shape[0]==0:
        return total_matrix
    len_new_matrix=new_matrix.shape[0]
    if index+len_new_matrix>total_matrix.shape[0]:
        temp_matrix=np.zeros((500000,5),dtype=float)
        total_matrix=np.vstack((total_matrix,temp_matrix))
    total_matrix[index:index+len_new_matrix]=new_matrix
    return total_matrix

def get_label_raw(fast5_fn, basecall_group, basecall_subgroup ,chromosome, position,strand):
    ##Open file
    # fast5_fn='/media/zhguo/QUO/ivt_positive/ivt_positive/single/42/878b9de7-043b-4d5e-b194-7771c5767ff5.fast5'
    # if '4a4438ea-fc46-42aa-9788-8d54f37471b9' not in fast5_fn:
    #     return None
    try:
        fast5_data = h5py.File(fast5_fn, 'r')
    except IOError:
        raise IOError('Error opening file. Likely a corrupted file.')

    # Get raw data
    try:
        raw_dat = list(fast5_data['/Raw/Reads/'].values())[0]
        raw_dat = raw_dat['Signal'][()]
    except:
        raise RuntimeError(
            'Raw data is not stored in Raw/Reads/Read_[read#] so ' +
            'new segments cannot be identified.')

    # Read corrected data
    try:
        corr_data = fast5_data['/Analyses/' + basecall_group + '/' + basecall_subgroup + '/Events']
        align_data = fast5_data['/Analyses/' + basecall_group + '/' + basecall_subgroup + '/Alignment']
        corr_attrs = dict(list(corr_data.attrs.items()))
        align_attrs=dict(list(align_data.attrs.items()))
        corr_data = corr_data[()]
    # ~ .value
    except:
        raise RuntimeError(('Corrected data not found.'))

    # fast5_info = fast5_data['UniqueGlobalKey/channel_id'].attrs
    # sampling_rate = fast5_info['sampling_rate'].astype('int_')

    # Reading extra information
    start_position=align_attrs['mapped_start']
    corr_start_rel_to_raw = corr_attrs['read_start_rel_to_raw']
    chrome=align_attrs["mapped_chrom"]
    if chrome != chromosome:
        return None
    if strand != align_attrs["mapped_strand"]:
        return None
    if position <= align_attrs['mapped_start'] or position >= align_attrs['mapped_end']:
        return None
    if len(raw_dat) > 99999999:
        raise ValueError(fast5_fn + ": max signal length exceed 99999999")
    if any(len(vals) <= 1 for vals in (corr_data, raw_dat)):
        raise NotImplementedError(('One or no segments or signal present in read.'))
    event_starts = corr_data['start'] + corr_start_rel_to_raw
    event_lengths = corr_data['length']
    event_bases = corr_data['base']
    fast5_data.close()
    # ~ label_data = np.array(
    # ~ list(zip(event_starts, event_lengths, event_bases)),
    # ~ dtype=[('start', '<u4'), ('length', '<u4'), ('base', 'S1')])
    return (raw_dat, event_bases, event_starts, event_lengths,start_position,strand,chrome)


def extract_feature(signal, event_start, event_length, base,start_position,end_position,strand,norm=True):
    global BASE_LIST
    position=FLAG.pos
    if strand == '+':
        current_index = position-start_position
    else:
        current_index = end_position - position - 1
    if current_index-FLAG.len < 0:
        start = 0
    else:
        start = current_index-FLAG.len
    if current_index + FLAG.len >= (end_position - start_position):
        end=len(event_start)
    else:
        end=current_index + FLAG.len + 1

    seq_array = np.array(list(base))
    assert seq_array.shape == event_length.shape
    # uniq_arr = np.unique(signal)
    # signal = (signal - np.median(uniq_arr)) / float(robust.mad(uniq_arr))
    # normalization

    signal = signal[event_start[0]:event_start[-1] + event_length[-1]]

    event_start = event_start-event_start[0]
    if norm:
        signal = normalize_signal_with_lim(signal)

    # filter too short or long dwell time
    # dwell_filter_pctls = (5, 95)
    # dwell_min, dwell_max = np.percentile(event_length, dwell_filter_pctls)
    # valid_bases = np.logical_and.reduce(
    #     (
    #         event_length > dwell_min,
    #         event_length < dwell_max,
    #         np.logical_not(np.isnan(event_length)),
    #     )
    # )
    # event_length[~valid_bases]=0


    total_feature_per_reads = []
    # if end >=event_start.shape[0]:
    #     #print(1)
    # draw_signal(norm_signal[event_start[start]:event_start[end]], event_start[start:end] - event_start[start],
    #             base[start:end])
    raw_signal_every = [signal[event_start[start + x]:event_start[start + x] + event_length[start + x]] for x in range(end-start)]

    for i, element in enumerate(raw_signal_every):
        if event_length[start + i] == 0:
            continue
        if strand == '+':
            final_position = position - FLAG.len + i
        else:
            final_position = position + FLAG.len - i
        temp = [np.mean(element), np.std(element), np.median(element), event_length[start + i], final_position]
        total_feature_per_reads.append(temp)
    total_feature_per_reads=np.array(total_feature_per_reads)
    return total_feature_per_reads

def extract_file(input_file,mode):
    basecall_group = FLAG.basecall_group
    basecall_subgroup = FLAG.basecall_subgroup

    chromosome = FLAG.chrom
    position = FLAG.pos
    strand = FLAG.strand
    try:
        temp_result = get_label_raw(input_file, basecall_group, basecall_subgroup,chromosome, position,strand)
        if temp_result is None:
            return None
        (raw_data, raw_label, raw_start, raw_length, start_position, strand,chrome) = temp_result
    except Exception as e:
        # print(str(e))
        return None

    # ~ print(input_file,raw_start,raw_length,raw_label)
    total_seq = "".join([x.decode() for x in raw_label])
    base_len=raw_label.shape[0]
    end_position = start_position+base_len
    if mode:
        raw_data = np.flip(raw_data)
    matrix_feature = extract_feature(raw_data, raw_start, raw_length, total_seq,start_position,end_position,strand,FLAG.norm)
    if matrix_feature is None:
        return None
    del raw_data

    return matrix_feature

def extract_group(args, total_fl,subsapmle_num=500):
    global FLAG,BASE_LIST
    BASE_LIST=None
    FLAG=args
    # feature_matrix = np.zeros((1000000,40))
    results_list = []
    ##########
    pool = multiprocessing.Pool(processes=int(args.cpu))
    ##########
    for fl in total_fl:
        result_per_read = pool.apply_async(extract_file, (fl,args.rna))
        results_list.append(result_per_read)
    pool.close()
    ############################
    pbar = tqdm(total=len(total_fl), position=0, leave=True)

    result_list=[]
    for result_per_read in results_list:
        temp= result_per_read.get()
        if temp is not None:
            feature_per_read = temp
            result_list.append(feature_per_read)

            del  feature_per_read
        pbar.update(1)
    #############################
    pool.join()
    pbar.close()

    num_aligned = len(result_list)
    if subsapmle_num < num_aligned:
        result_list=random.sample(result_list,subsapmle_num)
    final_feature=[]
    for item in result_list:
        final_feature.extend(item)
    df=pd.DataFrame(final_feature)


    if df.shape[0] == 0:
        raise Exception("can not find basecall_group or basecall_subgroup in fast5 files or there is no read aligned on the position")
    print('\nextracted ',num_aligned,' aligned reads from fast5 files')
    df.columns = ['Mean', 'STD', 'Median', 'Dwell time', 'position']

    # 转化为数值
    df['position'] = df['position'].astype(int).astype(str)

    if num_aligned > 50:
        dwell_filter_pctls = (0.5, 99.5)
        dwell_min, dwell_max = np.percentile(df['Dwell time'].values, dwell_filter_pctls)
        df = df[(df['Dwell time'] > dwell_min) & (df['Dwell time'] < dwell_max)]
        df.reset_index(inplace=True,drop=True)
    #     item_list = ['Mean', 'STD']
    #     for item in item_list:
    #         # collect data
    #         dwell_filter_pctls = (2.5, 97.5)
    #         dwell_min, dwell_max = np.percentile(df[item].values, dwell_filter_pctls)
    #         df = df[(df[item] > dwell_min) & (df[item] < dwell_max)]
    #         df.reset_index(inplace=True,drop=True)
    return df,num_aligned


def create_read_list_file(path,results_path):
    if path[-1] == '/':
        path = path[:-1]
    total_fl = []
    os.system("find " + path + " -name \"*.fast5\" >" + results_path + "/files.txt")
    read_name_list_file = results_path + '/files.txt'
    print("Generated fast5 list file: ", read_name_list_file)
    for i in open(read_name_list_file, "r"):
        total_fl.append(i.rstrip())
    cmd='rm '+read_name_list_file
    os.system(cmd)
    return total_fl


# if __name__ == '__main__':
#     parser = argparse.ArgumentParser()
#     # parser.add_argument('--basecall_group', default="RawGenomeCorrected_000",
#     #                     help='The attribute group to extract the training data from. e.g. RawGenomeCorrected_000')
#     # parser.add_argument('--basecall_subgroup', default='BaseCalled_template',
#     #                     help='Basecall subgroup Nanoraw resquiggle into. Default is BaseCalled_template')
#     # parser.add_argument('-i',"--fast5", default='/data/Ecoli_23s/data/L_rep2/single',
#     #                     help="fast5_file")
#     # parser.add_argument('-c',"--control_fast5", default='/data/Ecoli_23s/data/IVT_negative/single',
#     #                     help="control_fast5_file")
#     # parser.add_argument('-o',"--output", default="/data/Ecoli_23s/tombo_results_2030_plus", help="output_file")
#     # parser.add_argument("--chrom", default='NR_103073.1',help="Gene or chromosome name(head of your fasta file)")
#     # parser.add_argument("--pos", default=2030, type=int,help="site of your interest")
#     # parser.add_argument("--len", default=10, type=int, help="region around the position")
#     # parser.add_argument("--strand", default="+", help="Strand of your interest")
#     # parser.add_argument('-t',"--cpu", default=4, type=int, help="num of process")
#     # parser.add_argument("--ref", default="/data/Ecoli_23s/23S_rRNA.fasta", help="fasta file")
#     # parser.add_argument("--overplot-number", default=500, type=int, help="Number of read will be used to plot")
#     parser.add_argument('--basecall_group', default="RawGenomeCorrected_000",
#                         help='The attribute group to extract the training data from. e.g. RawGenomeCorrected_000')
#     parser.add_argument('--basecall_subgroup', default='BaseCalled_template',
#                         help='Basecall subgroup Nanoraw resquiggle into. Default is BaseCalled_template')
#     parser.add_argument('-i',"--fast5", required=True,
#                         help="fast5_file")
#     parser.add_argument('-c',"--control_fast5",
#                         help="control_fast5_file")
#     parser.add_argument('-o',"--output", default="tombo_result", help="output_file")
#     parser.add_argument("--chrom", required=True, help="Gene or chromosome name(head of your fasta file)")
#     parser.add_argument("--pos", required=True, type=int,help="site of your interest")
#     parser.add_argument("--len", default=10, type=int, help="region around the position")
#     parser.add_argument("--strand", default="+", help="Strand of your interest")
#     parser.add_argument('-t',"--cpu", default=4, type=int, help="num of process")
#     parser.add_argument("--ref", required=True, help="fasta file")
#     parser.add_argument("--overplot-number", default=500, type=int, help="Number of read will be used to plot")
#     args = parser.parse_args()
#     args.pos = args.pos - 1
#     subsapmle_num = args.overplot_number
#     fasta=read_fasta_to_dic(args.ref)
#     base_list = fasta[args.chrom][args.pos-args.len:args.pos+args.len+1]
#     if args.strand == '-':
#         base_list="".join(list(reversed(base_list)))
#         base_list=reverse_fasta(base_list)
#     results_path = args.output
#     if not os.path.exists(results_path):
#         os.mkdir(results_path)
#
#     title = args.chrom + ':' + str(args.pos - args.len + 1) + '-' + str(args.pos + args.len + 2) + ':' + args.strand
#
#     wt_file=create_read_list_file(args.fast5,results_path)
#     df_wt,aligned_num_wt=extract_group(args, wt_file,subsapmle_num)
#     df_wt['type']='Sample'
#     try:
#         ivt_file=create_read_list_file(args.control_fast5,results_path)
#         df_ivt,aligned_num_ivt=extract_group(args, ivt_file,subsapmle_num)
#         df_ivt['type'] = 'Control'
#         df=pd.concat([df_wt,df_ivt])
#         title = title + '   Sample:' + str(aligned_num_wt) + '  Control:' + str(aligned_num_ivt)
#         category = pd.api.types.CategoricalDtype(categories=['Sample', "Control"], ordered=True)
#         df['type'] = df['type'].astype(category)
#     except:
#         args.control = None
#     if args.control_fast5 is None:
#         df = df_wt
#         df_wt['type'] = 'Single'
#         title = title + '   Sample:' + str(aligned_num_wt)
#
#     category_data = [str(args.pos + x) for x in range(-args.len, args.len + 1)]
#     category = pd.api.types.CategoricalDtype(categories=category_data, ordered=True)
#     df['position'] = df['position'].astype(category)
#
#     # draw_volin(df,results_path,args.pos,base_list,title)
#     # draw_boxplot(df,results_path,args.pos,base_list,title)
#     signal_plot(df, results_path, args.pos, base_list, title, 'merged')
#     signal_plot(df, results_path, args.pos, base_list, title,'boxplot')
#     signal_plot(df, results_path, args.pos, base_list, title, 'violin_plot')
#     print('\nsaved as ', args.output)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="nanoCEM",
    version="0.0.2.0",
    author="GUO Zhihao",
    author_email="qhuozhihao@icloud.com",
    description='A simple tool designed to visualize the features that distinguish between two groups of ONT data at the site level.\
                It supports two re-squiggle program(tombo and f5c).',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lrslab/nanoCEM",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7,<3.10',
    install_requires=[
        'h5py==3.8.0',
        'numpy>=1.23.0',
        'pandas>=1.1.5',
        'plotnine>=0.8.0',
        'tqdm>=4.62.0',
        "pysam>=0.21.0",
        "pyslow5>=1.0.0",
        "vbz_h5py_plugin>=1.0.1"
    ],
    scripts=['nanoCEM/current_events_magnifier.py','nanoCEM/extract_sub_fast5_from_bam.py']
)


import numpy as np
OUTLIER_THRESH = 5.0

def normalize_signal_with_lim(signal, lower_lim=-OUTLIER_THRESH, upper_lim=OUTLIER_THRESH):
    shift = np.median(signal)
    # note factor to scale MAD to SD
    scale = np.median(np.absolute(signal - shift)) * 1.4826
    norm_signal = (signal - shift) / scale

    # norm_signal = c_apply_outlier_thresh(
    #         norm_signal, lower_lim, upper_lim)

    norm_signal = np.where(norm_signal > upper_lim, upper_lim, norm_signal)
    norm_signal = np.where(norm_signal < lower_lim, lower_lim, norm_signal)
    return norm_signal

def normalize_signal(signal):
    shift = np.median(signal)
    # note factor to scale MAD to SD
    scale = np.median(np.absolute(signal - shift)) * 1.4826
    norm_signal = (signal - shift) / scale

    # norm_signal = c_apply_outlier_thresh(
    #         norm_signal, lower_lim, upper_lim)
    return norm_signal


